/*
 * Copyright (c) 2002-2004 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.
 */

#define VERBOSE 0

/*
 * Sara Fleury
 * 
 * Segmentation des points laser. Utilise SEGKIT.
 * Appeler sickSegmentation.
 *
 */
#include <h2mathLib.h>
#include <stdio.h>
#include <string.h>
#ifdef VXWORKS
#include <vxWorks.h>
#include <sysLib.h>
#include <logLib.h>
#else
#include <math.h>
#include <portLib.h>
#endif /* VXWORKS */
#include <taskLib.h>
#include <errnoLib.h>
#include "h2mathLib.h"

/* sick module header */
#include "codels/sickNHeader.h"
#include "sickConst.h"

/* POM */
#include "pomStruct.h"

/* SEGKIT segmentation */
#include "lasdef.h"
#include "lasegm_p.h"
#include "las_struct_p.h"
#include "image_mgr_p.h"
#include "las_unc_p.h"

#include "sick_segs.h"
#include "sick_convert.h"

/* Intermediate image for segmentation */
static igla *ig=NULL;
static long numig=0;

/* Specific variance matrix for the LMS 200. Needs to be updated since the 
   values are from the laser at LAAS. */
static double var_pol[3] = {SICK_VAR_R,
			    SICK_VAR_RO,
			    SICK_VAR_O};

static BOOL getVariances;

static BOOL sickSegkitSegmentation(int iteration,
				   SICK_MEASURES_HEADER_STR *header, 
				   double rho[], 
				   SICK_SEG_PARAMS *segParams, 
				   SICK_SEGS_POSTER_STR *acqSegs);

static int filterPts(int *nbPts, p2d *pts);
static int reverseLonelyPts(int nbPts, SICK_POINT *pts);


/* ----------------------------------------------------------------------
 *
 *   sickAllocSickImage
 *
 *
 *
 */
static BOOL sickAllocSickImage (int nbPts)
{
  /* Already allocated -> should be free */
  if (ig != NULL) {
    /* Free image (sick points AND segments) */
    if (numig != -1) {
      Gfree_image(numig);
      numig = -1;
      ig = NULL;
    }
  }

  /* Allocate the image */
  numig = Get_Image();
  ig = PtoIm(numig);
  if(ig == NULL) {
    printf ("sickSegmentation: cannot allocate image\n");
    return FALSE;
  }

  /* Allocate array of points */
  ig->tbp = Pt_cree(nbPts);
  if (ig->tbp == NULL) {
    printf ("sickSegmentation: cannot allocate the %d points\n",
	    nbPts);
    return FALSE;
  }

  return TRUE;
}
				 
/* ----------------------------------------------------------------------
 *
 *   sickFillInSickImage - records the sick points for the segmentation
 *                         in the frame of the sensor
 *
 *
 */
static BOOL sickFillInSickImage(SICK_MEASURES_HEADER_STR *header, double rho[])
{
  double inc, alpha;
  int i, j=0;
  p2d *pt = ig->tbp; 

  ig->coosys = 'S';
  ig->x = 0.;
  ig->y = 0.;
  ig->cs = 1.;
  ig->sn = 0.;
  ig->cst = 1.;
  ig->snt = 0.;
  
  ig->lsgl = NULL;
  ig->nsg = 0;
  ig->nbp = 0;

  inc = header->dtheta;
  alpha =  header->theta_min;
  for (i = 0;  i < header->np; i++){

    if (rho[i] == 0)
      rho[i] = SICK_SEG_RHO_MAX;

    (pt+j)->r =  rho[i];

    /* !! ATTENTION !! : on n'enregistre pas tous les points 
       L'angle est donc important. */
    if((pt+j)->r < SICK_SEG_RHO_MAX){
      (pt+j)->o = alpha;
      (pt+j)->co = cos((pt+j)->o);
      (pt+j)->so = sin((pt+j)->o);
      (pt+j)->vr = var_pol[0];
      (pt+j)->vro = var_pol[1];
      (pt+j)->vo = var_pol[2];
      (pt+j)->x = (pt+j)->r * (pt+j)->co;
      (pt+j)->y = (pt+j)->r * (pt+j)->so;
      point_inc_roxy(pt+j);
      j += 1;
      ig->nbp++;
    }
    alpha += inc;
    
  }
  return numig;
}

/* ----------------------------------------------------------------------
 *
 *   sickFillInSickImage2   records the sick points for the segmentation
 *                          in the frame of the sensor
 *
 *
 */
static BOOL sickFillInSickImage2(int nbpts, SICK_POINT point[])
{
  int i;
  p2d *pt = ig->tbp; 

  ig->coosys = 'S';
  ig->x = 0.;
  ig->y = 0.;
  ig->cs = 1.;
  ig->sn = 0.;
  ig->cst = 1.;
  ig->snt = 0.;
  
  ig->lsgl = NULL;
  ig->nsg = 0;
  ig->nbp = 0;

  for (i = 0;  i < nbpts; i++){

    (pt+i)->r = hypot(point[i].x, point[i].y);
    (pt+i)->o = atan2(point[i].y, point[i].x);
    (pt+i)->co = cos((pt+i)->o);
    (pt+i)->so = sin((pt+i)->o);

    (pt+i)->vr = var_pol[0];
    (pt+i)->vro = var_pol[1];
    (pt+i)->vo = var_pol[2];
    (pt+i)->x = point[i].x;
    (pt+i)->y = point[i].y;

    if (i!=0 &&
	fabs((pt+i)->o  - (pt+i-1)->o) < .002) {
      printf ("BUG ? %d o %5.2f r %5.2f\n", i, (pt+i)->o, (pt+i)->r);
    }
    point_inc_roxy(pt+i);
  }
  ig->nbp = nbpts;
  return numig;
}

/*------------------------------------------------------------------------
 *
 * sickFillInSegs  - 
 *
 * Description: SEGKIT segs in SICK format
 * 
 * Retourne:    nb segs or -1 if problem
 */

static int sickFillInSegs(SICK_MEASURES_HEADER_STR *header, 
			  SICK_SEGS_POSTER_STR *acqSegs,
			  int nbOldSegs)
{
  PDBL *l;
  Sgl *s;
  int j;
  int nbSegs, nbLonely=0;
  int firstLonelyPt, lastLonelyPt;
  int nbSegsPts=0;

  /* Acquisition time */
  acqSegs->Header.sickTime = header->rcv_time;

  /* Raw data Frame */
  acqSegs->Header.frame = header->frame;

  /* Position */
  acqSegs->Header.sickPomPos = header->sickPomPos;

  /* Raw data Frame */
  acqSegs->Header.frame = header->frame;

  /* acquisition increment */
  acqSegs->Header.dtheta = header->dtheta;

  /* All segments */
  nbSegs = nbOldSegs;

  /*  !!! ATTENTION !!!! 
      Le nombre de points dans 'ig' peut-etre inferieur au nombre 
      de point acquis (car filtrage des points trop loins). 
      Il n'y a donc pas de correspondance directe entre le
      tableau de mesures du module et celui de SEGKIT (ig->tbp).
      Il faut donc aussi reprendre l'angle si on lit directement 'rho'
      dans 'tbp'.
      Enfin, les mesures dans 'tbp' sont dans le repere global.
  */
      

#if VERBOSE>=1
  printf ("---------  %d pts  ---------\n", ig->nbp);
#endif



  firstLonelyPt = ig->nbp - 1;
  for(l = ig->lsgl; l != NULL; l = dls(l)){
    s = (Sgl *)dlel(l);

    /* record segment */
    acqSegs->Segs.segs[nbSegs].x1 = s->first.x;
    acqSegs->Segs.segs[nbSegs].y1 = s->first.y;
    acqSegs->Segs.segs[nbSegs].x2 = s->last.x;
    acqSegs->Segs.segs[nbSegs].y2 = s->last.y;

#ifdef SEG_VARIANCES
    acqSegs->Sick_var_segs.var[nbSegs].r = s->r;
    acqSegs->Sick_var_segs.var[nbSegs].o = s->o;
    acqSegs->Sick_var_segs.var[nbSegs].vr = s->vr;
    acqSegs->Sick_var_segs.var[nbSegs].vo = s->vo;
    acqSegs->Sick_var_segs.var[nbSegs].vro = s->vro;
#endif

    lastLonelyPt = s->ilast + 1;
    nbSegsPts += (s->ilast - s->ifirst +1);

    /* record lonely points that are before the segment */
    for (j=firstLonelyPt; j>=lastLonelyPt; j--) {
      acqSegs->Points.point[nbLonely].x = ig->tbp[j].x;
      acqSegs->Points.point[nbLonely].y = ig->tbp[j].y;

    if (j!=ig->nbp - 1 &&
	fabs(ig->tbp[j+1].o  - ig->tbp[j].o) < .002) {
      printf ("BUG 1 ? %d o+1 %5.2f o %5.2f\n", 
	      j, ig->tbp[j+1].o, ig->tbp[j].o);
    }
#if VERBOSE>=3
      printf ("  lonely pt %d : j=%d o %5.2f x %5.2f y %5.2f r %5.2f\n",
	      nbLonely, j, ig->tbp[j].o, ig->tbp[j].x, ig->tbp[j].y, ig->tbp[j].r);
#endif
      nbLonely++;
    }
    firstLonelyPt = s->ifirst - 1;

    /* The segment coordinates */
#if VERBOSE>=2
    printf ("seg %5.2f %5.2f %5.2f %5.2f"
	    "   (pts %d to %d)\n",
	    acqSegs->Segs.segs[nbSegs].x1, 
	    acqSegs->Segs.segs[nbSegs].y1, 
	    acqSegs->Segs.segs[nbSegs].x2, 
	    acqSegs->Segs.segs[nbSegs].y2,
	    s->ifirst, s->ilast);
#endif
#if 0
    for (j=s->ilast; j>=s->ifirst; j--) {
      printf ("  seg pt %d : x %5.2f y %5.2f\n",
	      j, ig->tbp[j].x, ig->tbp[j].y);
    }
#endif

    /* next segment */
    nbSegs++;

    if (nbSegs>=SICK_MAX_ACQ_SEGS) {
      printf ("sickFillInSegs: too many segs. Keep only %d/%ld.\n",
	      SICK_MAX_ACQ_SEGS, ig->nsg);
      break;
    }
  }

  /* record lonely points that are after the last segment */
  for (j=firstLonelyPt; j>=0; j--) {
    acqSegs->Points.point[nbLonely].x = ig->tbp[j].x;
    acqSegs->Points.point[nbLonely].y = ig->tbp[j].y;
#if VERBOSE>=3
    printf ("  lonely pt %d : j=%d o %5.2f  x %5.2f y %5.2f\n",
	    nbLonely, j, ig->tbp[j].o, ig->tbp[j].x, ig->tbp[j].y);
#endif

    if (j!=ig->nbp - 1 &&
	fabs(ig->tbp[j+1].o  - ig->tbp[j].o) < .002) {
      printf ("BUG 2 ? %d o+1 %5.2f o %5.2f\n", 
	      j, ig->tbp[j+1].o, ig->tbp[j].o);
    }
    nbLonely++;
  }

  /* header */
  acqSegs->Header.nbLonelyPoints = nbLonely;
  if (nbSegs - nbOldSegs != ig->nsg) {
    printf ("sickFillInSegs: incoherent number of segs\n");
/*     return -1; */
  }
  acqSegs->Header.nbSegs = nbSegs;

#ifdef SEG_VARIANCES
  acqSegs->Header.variancesUpdated = SICK_TRUE;
#else 
  acqSegs->Header.variancesUpdated = SICK_FALSE;
#endif

#if VERBOSE>=1
/*   printf ("sickSeg: %d segs + %d pts\n", nbSegs, nbLonely); */
#endif

  reverseLonelyPts(nbLonely, acqSegs->Points.point);

  return nbSegs;
}


static int reverseLonelyPts(int nbPts, SICK_POINT *pts)
{
  SICK_POINT pt;
  int i;
  for(i=0; i<nbPts/2; i++) {
    pt = pts[i];
    pts[i] = pts[nbPts-1-i];
    pts[nbPts-1-i] = pt;
  }
  return nbPts;
}

/*------------------------------------------------------------------------
 *
 * sickUpdateSegsPoster  - exporte les segments acquis
 *
 * Description: from Copy_image_segment in image_mgr.c
 * 
 * Retourne:    nb segs or -1 if problem
 */

int sickUpdateSegsPoster(SICK_SEGS_POSTER_STR *acqSegs, POSTER_ID posterId)
{
  int size = sizeof(acqSegs->Header);
  int nb = acqSegs->Header.nbSegs;
  int nbLonely;

  size += nb * sizeof(SICK_SEG);

  /* Variances on segments' position */
#ifdef SEG_VARIANCES
  size -= nb * sizeof(SICK_SEG);
  size += SICK_MAX_ACQ_SEGS * sizeof(SICK_SEG) 
	  +  nb * sizeof(SICK_VAR_SEG);
#endif /* SEG_VARIANCES */

  /* update poster */
  if (posterWrite(posterId, 0, (void *)acqSegs, size) != size) {
    printf ("sickUpdateSegsPoster: posterWrite error\n");
    return -1;
  }
  
  nbLonely = acqSegs->Header.nbLonelyPoints;
  size = nbLonely * sizeof(SICK_POINT);
  if (posterWrite(posterId, 
		  sizeof(SICK_SEG_HEADER) + sizeof(SICK_SEGS),
		  (void *)acqSegs->Points.point, size) != size) {
    printf ("sickUpdateSegsPoster: posterWrite error\n");
    return -1;
  }
  
  return nb;
}


/* ----------------------------------------------------------------------
 *
 *   sickSegmentation
 *
 *
 *
 */

#define NB_RECOMMEND_PTS 6

BOOL sickSegmentation(SICK_MEASURES_HEADER_STR *header, 
		      double rho[], SICK_SEG_PARAMS *segParams, 
		      SICK_SEGS_POSTER_STR *acqSegs, POSTER_ID posterId, 
		      SICK_FRAME segExportFrame)
{
  int nbMinSegPts = segParams->nbMinSegPts;
  int nbMinSegPtsRequested = segParams->nbMinSegPts;
  int iter = 0;
  SICK_SEG_PARAMS segParamsLocal=*segParams;

  /**
   ** do the segmentation 
   **/

  if (nbMinSegPtsRequested < NB_RECOMMEND_PTS)
    nbMinSegPts = NB_RECOMMEND_PTS;

  /* do as many iterations as necessary in order first take the 
     segments with many points (at least NB_RECOMMEND_PTS)
     and to end with the required number of points. */
  do {
    segParamsLocal.nbMinSegPts = nbMinSegPts;
    if (!sickSegkitSegmentation(iter++, header, rho, &segParamsLocal, acqSegs)) 
      return FALSE;
#if VERBOSE>0
    printf ("sickSeg iter %d : %d seg (>= %d pts) %d pts\n", 
	    iter,
	    acqSegs->Header.nbSegs, nbMinSegPts, 
	    acqSegs->Header.nbLonelyPoints);
#endif
    nbMinSegPts--;
  } while (nbMinSegPts >= nbMinSegPtsRequested &&
	   acqSegs->Header.nbLonelyPoints > nbMinSegPts+1);

#if VERBOSE>1
  fprintf(stdout,"BEFORE CONVERT IN SEGMENTATION TO : %d\n",SDI_F->initParams.exportData.cartesianFrame);
  fprintf(stdout, "Robot Position : %f %f %f\n",
	  SDI_F->cartMeasures.header.sickPomPos.mainToOrigin.euler.x,
	  SDI_F->cartMeasures.header.sickPomPos.mainToOrigin.euler.y,
	  SDI_F->cartMeasures.header.sickPomPos.mainToOrigin.euler.z);

  
  int i=0;
  for (i=0;i < acqSegs->Header.nbSegs;i++) {
      fprintf(stdout,"i=%d (%lf,%lf) (%lf,%lf)\n",i,
	      acqSegs->Segs.segs[i].x1,
	      acqSegs->Segs.segs[i].y1,
	      acqSegs->Segs.segs[i].x2,
	      acqSegs->Segs.segs[i].y2);      
  }

#endif

  /**
   ** convert to required frame 
   **/
  sickConvertSICK_SEGSFrame(acqSegs, acqSegs, segExportFrame);

#if VERBOSE>1
  fprintf(stdout,"AFTER CONVERT IN SEGMENTATION TO : %d\n",SDI_F->initParams.exportData.cartesianFrame);
  fprintf(stdout, "Cartesian Segments Measures\n");
  fprintf(stdout, "Robot Position : %f %f %f\n",
	  SDI_F->cartMeasures.header.sickPomPos.mainToOrigin.euler.x,
	  SDI_F->cartMeasures.header.sickPomPos.mainToOrigin.euler.y,
	  SDI_F->cartMeasures.header.sickPomPos.mainToOrigin.euler.z);
  
  for (i=0;i < acqSegs->Header.nbSegs;i++) {
      fprintf(stdout,"i=%d (%lf,%lf) (%lf,%lf)\n",i,
	      acqSegs->Segs.segs[i].x1,
	      acqSegs->Segs.segs[i].y1,
	      acqSegs->Segs.segs[i].x2,
	      acqSegs->Segs.segs[i].y2);      
  }

#endif
  
  /**
   ** Save in poster 
   **/
  if (posterId != NULL) {
    if (sickUpdateSegsPoster(acqSegs, posterId) == -1) {
      return FALSE;
    }
  }
  return TRUE;
}


static BOOL sickSegkitSegmentation(int iteration,
				   SICK_MEASURES_HEADER_STR *header, 
				   double rho[], 
				   SICK_SEG_PARAMS *segParams, 
				   SICK_SEGS_POSTER_STR *acqSegs)
{
  int nbPts;
  int i;

  if (iteration == 0) 
    nbPts = header->np;
  else 
    nbPts = acqSegs->Header.nbLonelyPoints;

  if (nbPts <= 4) {
    printf ("SICK : only %d nbPts (iter %d)\n", nbPts, iteration);
    return FALSE;
  }

  /* (re)alloc */
  if (!sickAllocSickImage(nbPts)) {
    printf("sickSegmentation: alloc failed iter %d\n", iteration);
    return FALSE;
  }
  
  /* Fill-in sick image */
  if (iteration == 0) {
    acqSegs->Header.nbSegs = 0;
    sickFillInSickImage(header, rho);
  }
  else {
    sickFillInSickImage2(acqSegs->Header.nbLonelyPoints, 
			 acqSegs->Points.point);
  }

  if (ig->nbp <= 4) {
    printf ("SICK : only %d ig->nbpts (iter %d)\n", ig->nbp, iteration);
    return FALSE;
  }

#if 1
  /* XXXX temporarly for debug : test points validity */
  if (filterPts(&ig->nbp, ig->tbp)) {
    printf("SICK BUG iter %d : remove %d pts from %d (=%d) (remain %d)\n",
	   iteration, nbPts - ig->nbp, nbPts, 
	   acqSegs->Header.nbLonelyPoints, ig->nbp);
  }

  if (ig->nbp <= 4) {
    printf ("SICK : only %d ig->nbpts filtered (iter %d)\n", ig->nbp, iteration);
    return FALSE;
  }
#endif

  /* Set Segmentation params */
  conf_sg_host(segParams->leastSquareThres,
	       segParams->nbMinSegPts,
	       segParams->distMaxSegPts,
	       SICK_SEG_RHO_MAX);
  
  getVariances = segParams->getVariances;
  
  /* Segmentation */
  if (segmentation(PtoIm(numig),1) == -1) {
    printf ("sick_seg : segkit segmentation failed\n");
    for (i=0; i<acqSegs->Header.nbLonelyPoints; i++) {
      printf ("%d %g %g \n", i , 
	      acqSegs->Points.point[i].x,
	      acqSegs->Points.point[i].y);
    }
    return FALSE;
  }

  /* convert from segkit to sick format */
  sickFillInSegs(header, acqSegs, acqSegs->Header.nbSegs);

  return TRUE;
}

/* void set_encoders() */
/* { */
/*   printf ("sick_segs.c :  who uses set_encoders() ?\n"); */
/* } */


/* XXXX temporarly for debug : test points validity */
static int filterPts(int *nbPts, p2d *pts)
{
  int i, j;
  int first=0;
  int middle;
  double sens;
  int filter=0;

  if (*nbPts<=2) return 0;

  middle = (*nbPts)/2;
  sens = pts[middle+1].o - pts[middle].o;

  /* remove bad first pts */
  for (i=middle; i>0; i--) {
    if (sens > 0) {
      if (pts[i-1].o >= pts[i].o) {
	printf ("sick bug :remove pts 0 to %d, o =< %5.2f (keep from %d,  o >= %5.2f))\n", 
		i-1, RAD_TO_DEG(pts[i-1].o),
		i,  RAD_TO_DEG(pts[i].o));
	for (j=0; j<=i; j++) {
	  printf ("  pt %d, o = %5.2f  ",
		  j,  RAD_TO_DEG(pts[j].o));
	  printf ("x %5.2f y %5.2f r %5.2f\n", 
		  pts[j].x, 	pts[j].y, 	pts[j].r);
	}
	filter = 1;
	first = i;
	break;
      }
    }
    else {
      if (pts[i-1].o <= pts[i].o) {
	printf ("sick bug: remove pts 0 to %d, o =< %5.2f (keep from %d,  o <= %5.2f))\n", 
		i-1,  RAD_TO_DEG(pts[i-1].o),
		i,  RAD_TO_DEG(pts[i].o));
	for (j=0; j<=i; j++) {
	  printf ("  pt %d, o = %5.2f  ",
		  j,  RAD_TO_DEG(pts[j].o));
	  printf ("x %5.2f y %5.2f r %5.2f\n", 
		  pts[j].x, 	pts[j].y, 	pts[j].r);
	}
	filter = 1;
	first = i;
	break;
      }
    }
  }

  /* remove bad last pts */
  for (i=middle; i<*nbPts-1; i++) {
    if (sens > 0) {
      if (pts[i+1].o <= pts[i].o) {
	printf ("sick bug: remove pts %d to %d, o >= %5.2f (keep until %d,  o >= %5.2f))\n", 
		i+1, *nbPts-1,  RAD_TO_DEG(pts[i+1].o),
		i,  RAD_TO_DEG(pts[i].o));
	for (j=*nbPts-1; j>=i; j--) {
	  printf ("  pt %d, o = %5.2f  ",
		  j,  RAD_TO_DEG(pts[j].o));
	  printf ("x %5.2f y %5.2f r %5.2f\n", 
		  pts[j].x, 	pts[j].y, 	pts[j].r);
	}
	*nbPts = i+1;
	filter = 1;
	break;
      }
    }
    else {
      if (pts[i+1].o >= pts[i].o) {
	printf ("sick bug: remove pts %d to %d o >= %5.2f (keep until %d,  o <= %5.2f))\n", 
		i+1, *nbPts-1,  RAD_TO_DEG(pts[i+1].o),
		i,  RAD_TO_DEG(pts[i].o));
	for (j=*nbPts-1; j>=i; j--) {
	  printf ("  pt %d, o = %5.2f  ",
		  j,  RAD_TO_DEG(pts[j].o));
	  printf ("x %5.2f y %5.2f r %5.2f\n", 
		  pts[j].x, 	pts[j].y, 	pts[j].r);
	}
	*nbPts = i+1;
	filter = 1;
	break;
      }
    }
  }

  if (first) {
    for (i=first, j=0; i<*nbPts; i++, j++) {
      pts[j] = pts[i];
    }
  }
  *nbPts -= first;
  return filter;
}

