/*
 * Copyright (c) 2004 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.
 */

#ifndef UNIX
#include <vxWorks.h>
#else
#include <portLib.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "h2mathLib.h"

#include "pomStruct.h"
#include "libeuler/pomEuler.h"

#include "sickStruct.h"

#include "sick_convert.h"


/*------------------------------------------------------------------------
 * sickConvertSICK_SEGSFrame
 *
 * Convert SICK_SEGS_POSTER_STR from Header->frame to newFrame
 * 
 * sick_segs_in and sick_segs_out can refere to same data.
 */

BOOL sickConvertSICK_SEGSFrame (SICK_SEGS_POSTER_STR *sick_segs_in,
				SICK_SEGS_POSTER_STR *sick_segs_out,
				SICK_FRAME newFrame)
{
  sickConvertSegsFrame(&sick_segs_in->Header,
		       newFrame,
		       sick_segs_in->Segs.segs,
		       sick_segs_out->Segs.segs);

  sickConvertPtsFrame(sick_segs_in->Header.frame,
		      newFrame,
		      &sick_segs_in->Header.sickPomPos,
		      sick_segs_in->Header.nbLonelyPoints,
		      sick_segs_in->Points.point,
		      sick_segs_out->Points.point);

  if (sick_segs_out != sick_segs_in)
    sick_segs_out->Header = sick_segs_in->Header;
  sick_segs_out->Header.frame = newFrame;
  return TRUE;
}

/*------------------------------------------------------------------------
 * sickConvertSICK_CART_PTSFrame
 *
 * Convert SICK_SEGS_POSTER_STR from Header->frame to newFrame
 * 
 * sick_segs_in and sick_segs_out can refere to same data.
 */

BOOL sickConvertSICK_CART_PTSFrame (SICK_SCANCART_POSTER_STR *sick_cart_pts_in,
				    SICK_SCANCART_POSTER_STR *sick_cart_pts_out,
				    SICK_FRAME newFrame)
{
  sickConvertPtsFrame(sick_cart_pts_in->Header.frame,
		      newFrame,
		      &sick_cart_pts_in->Header.sickPomPos,
		      SICK_N_PT_MAX,
		      sick_cart_pts_in->Points,
		      sick_cart_pts_out->Points);

  if (sick_cart_pts_out != sick_cart_pts_in)
    sick_cart_pts_out->Header = sick_cart_pts_in->Header;
  sick_cart_pts_out->Header.frame = newFrame;
  return TRUE;
}

/*------------------------------------------------------------------------
 * sickConvertSegsFrame
 *
 * Convert segs coordinate from Header->frame to newFrame.
 *
 * segsIn and segsOut can refere to same data.
 * 
 * WARNING : be carfull to update Header->frame if required
 *           (not done by this function)
 *
 */

BOOL sickConvertSegsFrame (SICK_SEG_HEADER *Header,
			   SICK_FRAME newFrame,
			   SICK_SEG *segsIn,
			   SICK_SEG *segsOut)
{
  POM_EULER_V t3d;

  /* computes transformation */
  if (!sickGetT3DConvert(Header->frame, 
			 newFrame, 
			 &Header->sickPomPos,
			 &t3d)) {

    /* frame unchanged */
    if (segsIn != segsOut)
      memcpy(segsOut, segsIn, Header->nbSegs * sizeof(SICK_SEG));
    return TRUE;
  }

  /* convert */
  sickConvertSegs(&t3d, Header->nbSegs, segsIn, segsOut);
  return TRUE;

}


/*------------------------------------------------------------------------
 * sickConvertPtsFrame
 *
 * Convert points coordinate from Header->frame to newFrame.
 * ptsIn and ptsOut can refere to same data.
 *
 * WARNING : be carfull to update Header->frame if required
 *           (not done by this function)
 *
 */

BOOL sickConvertPtsFrame (SICK_FRAME oldFrame,
			  SICK_FRAME newFrame,
			  POM_SENSOR_POS *pos,
			  int nbPts,
			  SICK_POINT *ptsIn,
			  SICK_POINT *ptsOut)
{
  POM_EULER_V t3d;

  /* computes transformation */
  if (!sickGetT3DConvert(oldFrame, newFrame, 
			 pos, &t3d)) {

    /* frame unchanged */
    if (ptsIn != ptsOut)
      memcpy(ptsOut, ptsIn, nbPts * sizeof(SICK_POINT));
    return TRUE;
  }
  
  /* convert */
  sickConvertPts(&t3d, nbPts, ptsIn, ptsOut);
  return TRUE;

}


/*------------------------------------------------------------------------
 * sickPolarToCart1
 *
 * Convert polar measures to cartesian pts
 * See also sickPolarToCart2.
 *
 * WARNING : be carfull to update Header->frame if required
 *           (not done by this function)
 *
 */
void
sickPolarToCart1 (int nbPts, POM_SENSOR_POS *pos, SICK_FRAME cartFrame,
		  double *d, double *theta, SICK_POINT *sick_cp)
{
  double cosS, sinS;
  double cosR, sinR;
  double xs,ys;		/* coordonnee cartesiennes dans le repere sick */
  double xr,yr;		/* coordonnee cartesiennes dans le repere robot */
  int i;  

  cosS = cos(pos->sensorToMain.euler.yaw);
  sinS = sin(pos->sensorToMain.euler.yaw);

  cosR = cos(pos->mainToBase.euler.yaw);
  sinR = sin(pos->mainToBase.euler.yaw);

  /* for each measure */ /* XXXXXXX passer aux t3d */
  for (i = 0; i < nbPts; i++) {

    /* non null */
    if (*d != 0) {

      /* sick frame */
      xs = *d * cos(*theta);
      ys = *d * sin(*theta);
      if (cartFrame == SICK_SICK_FRAME) {
	sick_cp->x = xs;
	sick_cp->y = ys;
      }
      
      /* robot frame */
      else {
	xr = pos->sensorToMain.euler.x  +  xs * cosS - ys * sinS; 
	yr = pos->sensorToMain.euler.y  +  xs * sinS + ys * cosS;
	if (cartFrame == SICK_ROBOT_FRAME) {
	  sick_cp->x = xr;
	  sick_cp->y = yr;
	}
	
	/* global frame */
	else {
	  sick_cp->x = pos->mainToOrigin.euler.x +  xr * cosR - yr * sinR; 
	  sick_cp->y = pos->mainToOrigin.euler.y  +  xr * sinR + yr * cosR;
	}
      }
      /* next cartesian point */
      sick_cp++;
    }

    /* next measure */
    theta++;
    d++;
  } /* for */
}

/*------------------------------------------------------------------------
 * sickPolarToCart2
 *
 * Convert polar measures to cartesian pts
 * See also sickPolarToCart1.
 *
 * WARNING : be carfull to update Header->frame if required
 *           (not done by this function)
 *
 */
void
sickPolarToCart2 (int nbPts, POM_SENSOR_POS *pos, SICK_FRAME cartFrame,
		  double *d, double *thetaStart, double *dTheta, 
		  SICK_POINT *sick_cp)
{
  double cosS, sinS;
  double cosR, sinR;
  double xs,ys;		/* coordonnee cartesiennes dans le repere sick */
  double xr,yr;		/* coordonnee cartesiennes dans le repere robot */
  int i;  
  double theta = *thetaStart;

  cosS = cos(pos->sensorToMain.euler.yaw);
  sinS = sin(pos->sensorToMain.euler.yaw);

  cosR = cos(pos->mainToBase.euler.yaw);
  sinR = sin(pos->mainToBase.euler.yaw);

  /* for each measure */ /* XXXXXXX passer aux t3d */
  for (i = 0; i < nbPts; i++) {

    /* non null */
    if (*d != 0) {

      /* sick frame */
      xs = *d * cos(theta);
      ys = *d * sin(theta);
      if (cartFrame == SICK_SICK_FRAME) {
	sick_cp->x = xs;
	sick_cp->y = ys;
      }
      
      /* robot frame */
      else {
	xr = pos->sensorToMain.euler.x  +  xs * cosS - ys * sinS; 
	yr = pos->sensorToMain.euler.y  +  xs * sinS + ys * cosS;
	if (cartFrame == SICK_ROBOT_FRAME) {
	  sick_cp->x = xr;
	  sick_cp->y = yr;
	}
	
	/* global frame */
	else {
	  sick_cp->x = pos->mainToOrigin.euler.x +  xr * cosR - yr * sinR; 
	  sick_cp->y = pos->mainToOrigin.euler.y  +  xr * sinR + yr * cosR;
	}
      }
      /* next cartesian point */
      sick_cp++;
    }

    /* next measure */
    theta += *dTheta;
    d++;
  } /* for */
}


/*------------------------------------------------------------------------
 * sickGetT3DConvert
 *
 * Computes the trasnformation to be applied to convert from
 * Header->frame to newFrame.
 *
 * WARNING : be carfull to update Header->frame if required
 *           (not done by this function)
 *
 * Return FALSE if conversion is useless
 */

BOOL sickGetT3DConvert(SICK_FRAME oldFrame,
		       SICK_FRAME newFrame,
		       POM_SENSOR_POS *pos,
		       POM_EULER_V *t3d)
{
  switch (oldFrame)  {
  case SICK_SICK_FRAME:

    switch (newFrame) {
    case SICK_SICK_FRAME:
      /* sick -> sick : nothing to be done */
      return FALSE;

    case SICK_ROBOT_FRAME:
      /* sick -> robot */
      *t3d = pos->sensorToMain;
      return TRUE;

    case SICK_GLOBAL_FRAME:
      /* sick -> global */
      pomEulerVCompIn(t3d, &pos->sensorToMain, &pos->mainToOrigin);
      return TRUE;

    default:
      printf ("sickGetT3DConvert : unknown new frame %d\n", newFrame);
      return FALSE;
    }

    /* --  SICK_ROBOT_FRAME ------------------------------- */
  case SICK_ROBOT_FRAME:

    switch(newFrame) {
    case SICK_SICK_FRAME:
      /* robot -> sick */
      *t3d = pos->sensorToMain;
      pomEulerVInvert(t3d); 
      return TRUE;

    case SICK_ROBOT_FRAME:
      /* sick -> sick */
      return FALSE;
      
    case SICK_GLOBAL_FRAME:
      /* robot -> global */
      *t3d = pos->mainToOrigin;
      return TRUE;

    default:
      printf ("sickGetT3DConvert : unknown new frame %d\n", newFrame);
      return FALSE;
    }

    /* --  SICK_GLOBAL_FRAME ------------------------------- */
  case SICK_GLOBAL_FRAME :

    switch(newFrame) {
    case SICK_SICK_FRAME:
      /* global -> sick */
      pomEulerVCompIn(t3d, &pos->sensorToMain, &pos->mainToOrigin);
      pomEulerVInvert(t3d);
      return TRUE;

    case SICK_ROBOT_FRAME:
      /* origin -> robot */
      *t3d = pos->mainToOrigin;
      pomEulerVInvert(t3d);
      return TRUE;

    case SICK_GLOBAL_FRAME:
      return FALSE;

    default:
      printf ("sickGetT3DConvert : unknown new frame %d\n", newFrame);
      return FALSE;
    }

  default:
    printf ("sickGetT3DConvert : unknown old frame %d\n", 
	    oldFrame);
    return FALSE;
  }
}

/*------------------------------------------------------------------------
 *
 * sickConvertSegs  - 
 *
 * Description: convert seg frame unsing t3d transformation (2D !)
 * 
 * Retourne:    
 */

void sickConvertSegs(POM_EULER_V *t3d, int nbSegs, 
		     SICK_SEG *segsIn, SICK_SEG *segsOut)
{
  int i;
  SICK_SEG seg; /* intermediate in case segsIn == segsOut */
  double x0, y0, theta0, st, ct;

  x0 = t3d->euler.x;
  y0 = t3d->euler.y;
  theta0 = t3d->euler.yaw;
  ct = cos(theta0);
  st = sin(theta0);

  for (i=0; i<nbSegs;i++) {
    seg.x1 = x0 + segsIn[i].x1 * ct - segsIn[i].y1 * st;
    seg.y1 = y0 + segsIn[i].y1 * ct + segsIn[i].x1 * st;
    seg.x2 = x0 + segsIn[i].x2 * ct - segsIn[i].y2 * st;
    seg.y2 = y0 + segsIn[i].y2 * ct + segsIn[i].x2 * st;
    segsOut[i] = seg;
  }
}

/*------------------------------------------------------------------------
 *
 * sickConvertPts  - 
 *
 * Description: convert pt frame unsing t3d transformation (2D !)
 * ptsIn and ptsOut can refere to same data.
 * 
 * Retourne:    
 */

void sickConvertPts(POM_EULER_V *t3d, int nbPts, 
		    SICK_POINT *ptsIn, SICK_POINT *ptsOut)
{
  int i;
  SICK_POINT pt; /* intermediate in case ptsIn == ptsOut */
  double x0, y0, theta0, st, ct;

  x0 = t3d->euler.x;
  y0 = t3d->euler.y;
  theta0 = t3d->euler.yaw;
  ct = cos(theta0);
  st = sin(theta0);
  
  for (i=0; i<nbPts;i++) {
    pt.x = x0 + ptsIn[i].x * ct - ptsIn[i].y * st;
    pt.y = y0 + ptsIn[i].y * ct + ptsIn[i].x * st;
    ptsOut[i] = pt;
  }
}

