/**
 ** sike_save.c
 **
 ** Fonctions de sauvegarde de donnes du sick dans des fichiers
 ** et de re-lecture de ces donnes.
 **
 ** Auteur: Sara Fleury
 ** Date:   avril 2000
 **
 **/

#ifdef VXWORKS
#include <vxWorks.h>
#else
/* #include "genDecla.h"   */ /* Juste pour avoir BOOL, ... */
/* #include "portLib.h" */
#endif

#ifndef BOOL
#define BOOL int
#define TRUE 1
#define FALSE 0
#endif /* BOOL */

#include <stdio.h>
#include "h2mathLib.h"
#include "pomStruct.h"
#include "sickStruct.h"


/*------------------------------------------------------------------------
 *
 * sickSaveFile  --  sauvegarde des donnes dans le fichier <filename>
 *
 * Description: Donnes sauvegardes:
 *                . nombre de points, theta min, dtheta
 *                . position sick sur robot
 *                . odometrie lors de l'acquisition (x, y, theta, var[6], v, w)
 *                . les points (la distance rho)
 *
 * Retourne:    TRUE ou FALSE
 */

BOOL sickSaveFile (char *fileName, 
		   SICK_CONFIG_STR *config,
		   SICK_MEASURES_HEADER_STR *header,
		   double rho[SICK_N_PT_MAX])
{
  int i;
  FILE *file;
  double accuracy;

  /* Ouverture fichier */
  file = fopen(fileName, "w");
  if (file == NULL) {
    printf("sickSaveFile: cannot open %s\n", fileName);
    return FALSE;
  }

   switch(config->length_resolution) {
   case SICK_MEASURE_IN_MM:
     accuracy = 0.001;
     break;
   case SICK_MEASURE_IN_CM:
     accuracy = 0.01;
     break;
   case SICK_MEASURE_IN_10CM:
     accuracy = 0.1;
     break;
   default:
     printf ("InitSick: unknown measure sick length_resolution (default: 1 mm)\n");
     accuracy = 0.001;
     break;
   }

  /* Sauvegarde header */

   fprintf(file,
	   "date %lu\n"
	   "np %d  th-min %g dth %g\n"
	   "sick  x %g y %g th %g acc %g max %d\n"
	   "robot x %g y %g th %g\n"
	   " var %g %g %g %g %g %g\n"
	   " vx %g vy %g w %g\n",
	   header->rcv_time,
	   header->np, header->theta_min, header->dtheta,

	   header->sickPomPos.sensorToMain.euler.x,
	   header->sickPomPos.sensorToMain.euler.y,
	   header->sickPomPos.sensorToMain.euler.yaw,

	   accuracy, config->length_range_max,

	   header->sickPomPos.mainToBase.euler.x,
	   header->sickPomPos.mainToBase.euler.y,
	   header->sickPomPos.mainToBase.euler.yaw,
	   
     
	   header->sickPomPos.mainToBase.var.vxx,
	   header->sickPomPos.mainToBase.var.vxy,
	   header->sickPomPos.mainToBase.var.vyy,
	   header->sickPomPos.mainToBase.var.vox,
	   header->sickPomPos.mainToBase.var.voy,
	   header->sickPomPos.mainToBase.var.voo,
	   
	   header->sickPomPos.VLocal.euler.x,
	   header->sickPomPos.VLocal.euler.y,
	   header->sickPomPos.VLocal.euler.yaw);


   /* Sauvegarde mesures */
   for (i=0; i<header->np;) {
     fprintf(file, "%g ", rho[i]);
     i++;
     if (i%5==0)
       fprintf(file, "\n");
   }

   /* Fermeture fichier */
   fclose(file);
   
   return TRUE;
}

/*------------------------------------------------------------------------
 *
 * sickReadFile  --  lecture des donnes dans le fichier <filename>
 *
 * Description: Donnes lues:
 *                . nombre de points, theta min, dtheta
 *                . position sick sur robot
 *                . odometrie lors de l'acquisition (x, y, theta, var[6], v, w)
 *                . les points (la distance rho)
 *
 * Retourne:    TRUE ou FALSE
 */

BOOL sickReadFile (char *fileName, 
		   SICK_CONFIG_STR *config,
		   SICK_MEASURES_HEADER_STR *header,
		   double rho[SICK_N_PT_MAX])
{
  int i;
  FILE *file;
  double accuracy;
  int length_range;
  char test;
  /* Ouverture fichier */
  file = fopen(fileName, "r");
  if (file == NULL) {
    printf("sickReadFile: cannot open %s\n", fileName);
    return FALSE;
  }

  /* Lecture header */
  if (header == NULL) {
    printf("sickReadFile: header is null!\n");
    fclose(file);
    return FALSE;
  }

  fscanf(file,
	 "np %d  th-min %lf dth %lf\n"
	 "sick  x %lf y %lf th %lf acc %lf max %d\n"
	 "robot x %lf y %lf th %lf\n"
	 " var %lf %lf %lf %lf %lf %lf\n",
	 &header->np, &header->theta_min, &header->dtheta,
	 
	 
	 &header->sickPomPos.sensorToMain.euler.x,
	 &header->sickPomPos.sensorToMain.euler.y,
	 &header->sickPomPos.sensorToMain.euler.yaw,
	 
	 &accuracy, &length_range,
	 
	 &header->sickPomPos.mainToBase.euler.x,
	 &header->sickPomPos.mainToBase.euler.y,
	 &header->sickPomPos.mainToBase.euler.yaw,
	   
     
	 &header->sickPomPos.mainToBase.var.vxx,
	 &header->sickPomPos.mainToBase.var.vxy,
	 &header->sickPomPos.mainToBase.var.vyy,
	 &header->sickPomPos.mainToBase.var.vox,
	 &header->sickPomPos.mainToBase.var.voy,
	 &header->sickPomPos.mainToBase.var.voo
	 );
  /**
   * For old format compliance
   *
   */
  test=getc(file);
  test=getc(file);
  if (test=='x') {
    /* fprintf(stderr,"NEW FORMAT !\n"); */
    fscanf(file,
	   "%lf vy %lf w %lf\n",
	   &header->sickPomPos.VLocal.euler.x,
	   &header->sickPomPos.VLocal.euler.y,
	   &header->sickPomPos.VLocal.euler.yaw);
  }
  else {
    /* fprintf(stderr,"OLD FORMAT !\n"); */
    fscanf(file,
	   "%lf w %lf\n",
	   &header->sickPomPos.VLocal.euler.x,
	   &header->sickPomPos.VLocal.euler.yaw);
    header->sickPomPos.VLocal.euler.y=0.0;
  }

	   /* Config */
  if (config != NULL) {
    config->length_range_max = length_range;

    if (accuracy == 0.001)
      config->length_resolution = SICK_MEASURE_IN_MM;
    else if (accuracy == 0.01)
      config->length_resolution = SICK_MEASURE_IN_CM;
    else if (accuracy == 0.1)
      config->length_resolution = SICK_MEASURE_IN_10CM;
    else {
      printf ("InitSick: unknown measure sick length_resolution (default: 1 mm)\n");
      config->length_resolution = SICK_MEASURE_IN_MM;
    }
    
    config->angle_range = RAD_TO_DEG(header->np*header->dtheta) < 140 ?
		    SICK_SCAN_100_DEG : SICK_SCAN_180_DEG;

    config->angle_resolution = header->dtheta > DEG_TO_RAD(0.7) ?
			 SICK_SCAN_ONE_DEG_RESOLUTION :
			   SICK_SCAN_HALF_DEG_RESOLUTION;
  }

  /* Lecture mesures */
  if (rho != NULL) {
    for (i=0; i<header->np;) {
      fscanf(file, "%lf ", &rho[i]);
      i++;
      if (i%5==0)
	fscanf(file, "\n");
    }
  }

  /* Fermeture fichier */
  fclose(file);

  return TRUE;
}
